import math
import scipy.stats as stats
from scipy.interpolate import interp1d
from .uk_who import uk_who_lms_array_for_measurement_and_sex
from .turner import turner_lms_array_for_measurement_and_sex
from .trisomy_21 import trisomy_21_lms_array_for_measurement_and_sex
# from scipy import interpolate  #see below, comment back in if swapping interpolation method
# from scipy.interpolate import CubicSpline #see below, comment back in if swapping interpolation method
from .constants.reference_constants import UK_WHO, TURNERS, TRISOMY_21, COLE_TWO_THIRDS_SDS_NINE_CENTILES, COLE_TWO_THIRDS_SDS_NINE_CENTILE_COLLECTION, THREE_PERCENT_CENTILE_COLLECTION, MEASUREMENT_METHODS, SEXES, UK_WHO_REFERENCES
import logging
import json
import pkg_resources


def cubic_interpolation(age: float, age_one_below: float, age_two_below: float, age_one_above: float, age_two_above: float, parameter_two_below: float, parameter_one_below: float, parameter_one_above: float, parameter_two_above: float) -> float:
    """
    See sds function. This method tests if the age of the child (either corrected for prematurity or chronological) is at a threshold of the reference data
    This method is specific to the UK-WHO data set.
    """

    cubic_interpolated_value = 0.0

    t = 0.0  # actual age ///This commented function is Tim Cole"s used in LMSGrowth to perform cubic interpolation - 50000000 loops, best of 5: 7.37 nsec per loop
    tt0 = 0.0
    tt1 = 0.0
    tt2 = 0.0
    tt3 = 0.0

    t01 = 0.0
    t02 = 0.0
    t03 = 0.0
    t12 = 0.0
    t13 = 0.0
    t23 = 0.0

    t = age

    tt0 = t - age_two_below
    tt1 = t - age_one_below
    tt2 = t - age_one_above
    tt3 = t - age_two_above

    t01 = age_two_below - age_one_below
    t02 = age_two_below - age_one_above
    t03 = age_two_below - age_two_above

    t12 = age_one_below - age_one_above
    t13 = age_one_below - age_two_above
    t23 = age_one_above - age_two_above

    cubic_interpolated_value = parameter_two_below * tt1 * tt2 * tt3 / t01 / t02 / t03 - parameter_one_below * tt0 * tt2 * tt3 / t01 / \
        t12 / t13 + parameter_one_above * tt0 * tt1 * tt3 / t02 / t12 / \
        t23 - parameter_two_above * tt0 * tt1 * tt2 / t03 / t13 / t23

    # prerequisite arrays for either of below functions
    # xpoints = [age_two_below, age_one_below, age_one_above, age_two_above]
    # ypoints = [parameter_two_below, parameter_one_below, parameter_one_above, parameter_two_above]

    # this is the scipy cubic spline interpolation function...
    # cs = CubicSpline(xpoints,ypoints,bc_type="natural")
    # cubic_interpolated_value = cs(age) # this also works, but not as accurate: 50000000 loops, best of 5: 7.42 nsec per loop

    # this is the scipy splrep function
    # tck = interpolate.splrep(xpoints, ypoints)
    # cubic_interpolated_value = interpolate.splev(age, tck)   #Matches Tim Cole"s for accuracy but slower: speed - 50000000 loops, best of 5: 7.62 nsec per loop

    return cubic_interpolated_value


def linear_interpolation(age: float, age_one_below: float, age_one_above: float, parameter_one_below: float, parameter_one_above: float) -> float:
    """
    See sds function. This method is to do linear interpolation of L, M and S values for children whose ages are at the threshold of the reference data, making cubic interpolation impossible
    """

    linear_interpolated_value = 0.0

    # linear_interpolated_value = parameter_one_above + (((decimal_age - age_below)*parameter_one_above-parameter_one_below))/(age_above-age_below)
    x_array = [age_one_below, age_one_above]
    y_array = [parameter_one_below, parameter_one_above]
    intermediate = interp1d(x_array, y_array)
    linear_interpolated_value = intermediate(age)
    return linear_interpolated_value


def z_score(l: float, m: float, s: float, observation: float):
    """
    Converts the (age-specific) L, M and S parameters into a z-score
    """
    sds = 0.0
    if l != 0.0:
        sds = (((math.pow((observation / m), l)) - 1) / (l * s))
    else:
        sds = (math.log(observation / m) / s)
    return sds


def centile(z_score: float):
    """
    Converts a Z Score to a p value (2-tailed) using the SciPy library, which it returns as a percentage
    """
    try:
        centile = (stats.norm.cdf(z_score) * 100)
        return centile
    except TypeError as err:
        raise TypeError(err)


def measurement_for_z(z: float, l: float, m: float, s: float) -> float:
    """
    Returns a measurement for a z score, L, M and S
    """
    if l != 0.0:
        measurement_value = math.pow((1 + l * s * z), 1 / l) * m
    else:
        measurement_value = math.exp(s * z) * m
    return measurement_value


def nearest_lowest_index(
    lms_array: list,
    age: float
) -> int:
    """
    loops through the array of LMS values and returns either 
    the index of an exact match or the lowest nearest decimal age
    """
    lowest_index = 0
    for num, lms_element in enumerate(lms_array):
        reference_age = lms_element["decimal_age"]
        if round(reference_age, 16) == round(age, 16):
            lowest_index = num
            break
        else:
            if lms_element["decimal_age"] < age:
                lowest_index = num
    return lowest_index


def fetch_lms(age: float, lms_value_array_for_measurement: list):
    """
    Retuns the LMS for a given age. If there is no exact match in the reference
    an interpolated LMS is returned. Cubic interpolation is used except at the fringes of the 
    reference where linear interpolation is used.
    It accepts the age and a python list of the LMS values for that measurement_method and sex.
    """
    age_matched_index = nearest_lowest_index(
        lms_value_array_for_measurement, age)  # returns nearest LMS for age
    if round(lms_value_array_for_measurement[age_matched_index]["decimal_age"], 16) == round(age, 16):
        # there is an exact match in the data with the requested age
        l = lms_value_array_for_measurement[age_matched_index]["L"]
        m = lms_value_array_for_measurement[age_matched_index]["M"]
        s = lms_value_array_for_measurement[age_matched_index]["S"]
    else:
        # there has not been an exact match in the reference data
        # Interpolation will be required.
        # The age_matched_index is one below the age supplied. There
        # needs to be a value below that, and two values above the supplied age,
        # for cubic interpolation to be possible.
        age_one_below = lms_value_array_for_measurement[age_matched_index]["decimal_age"]
        age_one_above = lms_value_array_for_measurement[age_matched_index + 1]["decimal_age"]
        parameter_one_below = lms_value_array_for_measurement[age_matched_index]
        parameter_one_above = lms_value_array_for_measurement[age_matched_index + 1]

        if age_matched_index >= 1 and age_matched_index < len(lms_value_array_for_measurement) - 2:
            # cubic interpolation is possible
            age_two_below = lms_value_array_for_measurement[age_matched_index - 1]["decimal_age"]
            age_two_above = lms_value_array_for_measurement[age_matched_index + 2]["decimal_age"]
            parameter_two_below = lms_value_array_for_measurement[age_matched_index - 1]
            parameter_two_above = lms_value_array_for_measurement[age_matched_index + 2]

            l = cubic_interpolation(age=age, age_one_below=age_one_below, age_two_below=age_two_below, age_one_above=age_one_above, age_two_above=age_two_above,
                                    parameter_two_below=parameter_two_below["L"], parameter_one_below=parameter_one_below["L"], parameter_one_above=parameter_one_above["L"], parameter_two_above=parameter_two_above["L"])
            m = cubic_interpolation(age=age, age_one_below=age_one_below, age_two_below=age_two_below, age_one_above=age_one_above, age_two_above=age_two_above,
                                    parameter_two_below=parameter_two_below["M"], parameter_one_below=parameter_one_below["M"], parameter_one_above=parameter_one_above["M"], parameter_two_above=parameter_two_above["M"])
            s = cubic_interpolation(age=age, age_one_below=age_one_below, age_two_below=age_two_below, age_one_above=age_one_above, age_two_above=age_two_above,
                                    parameter_two_below=parameter_two_below["S"], parameter_one_below=parameter_one_below["S"], parameter_one_above=parameter_one_above["S"], parameter_two_above=parameter_two_above["S"])
        else:
            # we are at the thresholds of this reference. Only linear interpolation is possible
            l = linear_interpolation(age=age, age_one_below=age_one_below, age_one_above=age_one_above,
                                     parameter_one_below=parameter_one_below["L"], parameter_one_above=parameter_one_above["L"])
            m = linear_interpolation(age=age, age_one_below=age_one_below, age_one_above=age_one_above,
                                     parameter_one_below=parameter_one_below["M"], parameter_one_above=parameter_one_above["M"])
            s = linear_interpolation(age=age, age_one_below=age_one_below, age_one_above=age_one_above,
                                     parameter_one_below=parameter_one_below["S"], parameter_one_above=parameter_one_above["S"])

    return {
        "l": l,
        "m": m,
        "s": s
    }


def measurement_from_sds(
    reference: str,
    requested_sds: float,
    measurement_method: str,
    sex: str,
    age: float,
    born_preterm: bool = False
) -> float:

    try:
        lms_value_array_for_measurement = lms_value_array_for_measurement_for_reference(
            reference=reference, age=age, measurement_method=measurement_method, sex=sex, born_preterm=born_preterm)
    except LookupError as err:
        raise LookupError(err)

    # get LMS values from the reference: check for age match, interpolate if none
    lms = fetch_lms(
        age=age, lms_value_array_for_measurement=lms_value_array_for_measurement)
    l = lms["l"]
    m = lms["m"]
    s = lms["s"]

    observation_value = measurement_for_z(z=requested_sds, l=l, m=m, s=s)
    return observation_value


def sds_for_measurement(
    reference: str,
    age: float,
    measurement_method: str,
    observation_value: float,
    sex: str,
    born_preterm: bool = False
) -> float:

    try:
        lms_value_array_for_measurement = lms_value_array_for_measurement_for_reference(
            reference=reference, age=age, measurement_method=measurement_method, sex=sex, born_preterm=born_preterm)
    except LookupError as err:
        raise LookupError(err)

    # get LMS values from the reference: check for age match, interpolate if none
    lms = fetch_lms(
        age=age, lms_value_array_for_measurement=lms_value_array_for_measurement)
    l = lms["l"]
    m = lms["m"]
    s = lms["s"]

    return z_score(l=l, m=m, s=s, observation=observation_value)


def percentage_median_bmi(reference: str, age: float, actual_bmi: float, sex: str, born_preterm=False) -> float:
    """
    public method
    This returns a child"s BMI expressed as a percentage of the median value for age and sex.
    It is used widely in the assessment of malnutrition particularly in children and young people with eating disorders.
    It accepts the reference ('uk-who', 'turners-syndrome' or 'trisomy-21')
    """

    # fetch the LMS values for the requested measurement
    try:
        lms_value_array_for_measurement = lms_value_array_for_measurement_for_reference(
            reference=reference, measurement_method="bmi", sex=sex, age=age, born_preterm=born_preterm)
    except LookupError as err:
        raise LookupError(err)

    # get LMS values from the reference: check for age match, interpolate if none
    try:
        lms = fetch_lms(
            age=age, lms_value_array_for_measurement=lms_value_array_for_measurement)
    except LookupError as err:
        print(err)
        return None

    m = lms["m"]  # this is the median BMI

    percent_median_bmi = (actual_bmi / m) * 100.0
    return percent_median_bmi


def generate_centile(z: float, centile: float, measurement_method: str, sex: str, lms_array_for_measurement: list, reference: str) -> list:
    """
    Generates a centile curve for a given reference. 
    Takes the z-score equivalent of the centile, the centile to be used as a label, the sex and measurement method.
    """

    min_age = lms_array_for_measurement[0]["decimal_age"]
    max_age = lms_array_for_measurement[-1]["decimal_age"]

    centile_measurements = []
    age = min_age
    while age <= max_age:
        # loop through the reference in steps of 0.1y
        try:
            measurement = measurement_from_sds(
                reference=reference, measurement_method=measurement_method, requested_sds=z, sex=sex, age=age, born_preterm=True)
        except Exception as err:
            print(err)
            measurement = None

        # creates a data point
        if measurement is not None:
            rounded = round(measurement, 4)
        else:
            rounded = None
        value = {
            "l": centile,
            "x": round(age, 4),
            "y": rounded
        }

        centile_measurements.append(value)

        # weekly intervals until 2 y, then monthly
        if age <= 2:
            age += (7 / 365.25)  # weekly intervals
        else:
            age += 1 / 12  # monthly intervals

        # Although it is preferable to have weekly data points, it generates files of ~2.5 MB
        # even after minifying, which are not practical. Weekly values makes plotting easier.
        # Here we have used weekly points from preterm to 2 y, monthly values after.
        # age += (7/365.25) # weekly intervals
    return centile_measurements


def rounded_sds_for_centile(centile: float) -> float:
    """
    converts a centile (supplied as a percentage) using the scipy package to the nearest 2/3 SDS.
    """
    sds = stats.norm.ppf(centile / 100)
    if sds == 0:
        return sds
    else:
        rounded_to_nearest_two_thirds = round(sds / (2 / 3))
        return rounded_to_nearest_two_thirds * (2 / 3)


def sds_for_centile(centile: float) -> float:
    """
    converts a centile (supplied as a percentage) using the scipy package to an SDS.
    """
    sds = stats.norm.ppf(centile / 100)
    return sds


def lms_value_array_for_measurement_for_reference(
    reference: str,
    age: float,
    measurement_method: str,
    sex: str,
    born_preterm: bool
) -> list:
    """
    This is a private function which returns the LMS array for measurement_method and sex and reference
    It accepts the reference ('uk-who', 'turners-syndrome' or 'trisomy-21')
    """

    if reference == UK_WHO:
        try:
            lms_value_array_for_measurement = uk_who_lms_array_for_measurement_and_sex(
                age=age, measurement_method=measurement_method, sex=sex, born_preterm=born_preterm)
        except LookupError as error:
            raise LookupError(error)
    elif reference == TURNERS:
        try:
            lms_value_array_for_measurement = turner_lms_array_for_measurement_and_sex(
                measurement_method=measurement_method, sex=sex, age=age)
        except LookupError as error:
            raise LookupError(error)
    elif reference == TRISOMY_21:
        try:
            lms_value_array_for_measurement = trisomy_21_lms_array_for_measurement_and_sex(
                measurement_method=measurement_method, sex=sex, age=age)
        except LookupError as error:
            raise LookupError(error)
    else:
        raise ValueError("Incorrect reference supplied")
    return lms_value_array_for_measurement



import json
import pkg_resources
from .constants import *
# import timeit #see below, comment back in if timing functions in this module

"""
birth_date: date of birth
observation_date: date of observation
sex: sex (string, MALE or FEMALE)
decimal_age: chronological, decimal
corrected_age: corrected for prematurity, decimal
measurement_method: height, weight, bmi, ofc (decimal)
observation: value (float)
gestation_weeks: gestational age(weeks), integer
gestation_days: supplementary days of gestation
lms: L, M or S
reference: reference data
"""

#load the reference data

TURNER_DATA = pkg_resources.resource_filename(__name__, "/data_tables/turner.json")
with open(TURNER_DATA) as json_file:
            TURNER_DATA = json.load(json_file)
            json_file.close()

def turner_lms_array_for_measurement_and_sex(
        measurement_method: str,    
        sex: str,  
        age: float
    ):

    invalid_data, data_error = reference_data_absent(age=age, measurement_method=measurement_method, sex=sex)

    if invalid_data:
        raise LookupError(data_error)

    # Get the Turner reference data
    try:
        return TURNER_DATA["measurement"][measurement_method][sex]
    except: # there is no reference for the age supplied
        raise LookupError("The Turner's syndrome reference cannot be found.")
    

def reference_data_absent( 
        age: float,
        measurement_method: str,
        sex: str
    ):
    """
    Helper function.
    Returns boolean
    Tests presence of valid reference data for a given measurement request

    Turners syndrome
    Data are available for girls heights at year intervals from 1-20y
    No other reference data are available
    """

    if age < 1: # lower threshold of Turner data
        return True, 'There is no reference data below 1 year.'
    elif age > TWENTY_YEARS: # upper threshold of Turner data
        return True, "There is no reference data above 20 years."
    elif measurement_method=="weight" or measurement_method=="bmi" or measurement_method=="ofc":
        text_string = measurement_method
        if measurement_method == "bmi":
            text_string = "BMI (body mass index)"
        elif measurement_method == "ofc":
            text_string = "head circumference"
        return True, f"There is no reference data for {text_string}."
    elif sex == "male":
        return True, f"Turner's syndrome only affects girls and women."
    else:
        return False, "Valid Data"

def select_reference_data_for_turner(measurement_method: str, sex: str):
    return turner_lms_array_for_measurement_and_sex(measurement_method=measurement_method, sex=sex, age=1.0)


from .date_calculations import chronological_decimal_age, corrected_decimal_age, chronological_calendar_age, estimated_date_delivery, corrected_gestational_age
from .global_functions import centile, sds_for_measurement, measurement_from_sds, percentage_median_bmi
from .centile_bands import centile_band_for_centile
from .bmi_functions import bmi_from_height_weight, weight_for_bmi_height
from .growth_interpretations import comment_prematurity_correction
from .fictional_child import generate_fictional_child_data
from .dynamic_growth import correlate_weight, create_correlation_surface, create_thrive_lines
from .uk_who import select_reference_data_for_uk_who_chart
from .turner import select_reference_data_for_turner
from .trisomy_21 import select_reference_data_for_trisomy_21
from .measurement import Measurement
from .chart_functions import create_chart, create_plottable_child_data
from .constants import *



# from .measurement import Measurement
# import pandas as pd
import numpy as np
import pandas as pd
# import matplotlib.pyplot as plt
import os
import math
from scipy.interpolate import interp1d
from datetime import date
from dateutil.relativedelta import relativedelta
from random import uniform
from .measurement import Measurement
from .global_functions import measurement_from_sds
from .date_calculations import corrected_decimal_age

"""
These functions are experimental
Height, weight, BMI or OFC in terms of SDS / Centile are snapshots in time and tell
us actually very little about growth, which is a dynamic measure. In order to make 
predictions, we need to look at change in parameter measured over time (velocity)
which requires 2 measurements over a known time interval, or change in velocity (acceleration/deceleration)
which requires three measurements. 

From these measurements predictions can be made about speed of growth, or rate of slowing (catch down)
or acceleration (catch up). The normative data against which to compare the index child are 
thrive lines, generated here.

"""


def velocity(parameter: str, measurements_array):
    """
    This is an experimental function and not to be used clinically because velocity is not constant
    and is age dependent.
    Velocity needs at least 2 measurements from 2 consecutive time points.
    This takes an array of Measurement objects of the same child, removes the last 2 values of the same
    measurement and calculates the velocity in units/y.
    """
    parameter_list = []
    if len(measurements_array) < 2:
        return 'Not enough data'
    else:
        for measurement in measurements_array:
            if measurement:
                if parameter == measurement['child_observation_value']['measurement_method']:
                    parameter_list.append(measurement)
        if len(parameter_list) < 2:
            return f"There are not enough {parameter} values to calculate a velocity."
        else:
            last = parameter_list[-1]
            penultimate = parameter_list[-2]
            time_elapsed = last['measurement_dates']['chronological_decimal_age'] - \
                penultimate['measurement_dates']['chronological_decimal_age']
            parameter_difference = 0.0
            parameter_difference = last['child_observation_value']['observation_value'] - \
                penultimate['child_observation_value']['observation_value']
            return parameter_difference / time_elapsed


def acceleration(parameter: str, measurements_array):
    """
    This is an experimental function and not to be used clinically because acceleration is not constant
    and is age dependent.
    Accelaration needs at least 3 measurements over 3 consecutive time points in order to compare the velocity
    change between the first pair and the last pair.
    This takes an array of Measurement objects of the same child, removes the last 3 values of the same
    measurement and calculates the acceleration.
    The parameter in question is one of 'height', 'weight', 'bmi', 'ofc'
    """
    parameter_list = []
    if len(measurements_array) < 3:
        return 'Not enough data'
    else:
        for measurement in measurements_array:
            if measurement:
                if parameter == measurement['child_observation_value']['measurement_method']:
                    parameter_list.append(measurement)
        if len(parameter_list) < 3:
            return f"There are not enough {parameter} values to calculate acceleration."
        else:
            last = parameter_list[-1]
            penultimate = parameter_list[-2]
            antepentultimate = parameter_list[-3]
            first_parameter_pair_time_elapsed = penultimate['measurement_dates']['chronological_decimal_age'] - \
                antepentultimate['measurement_dates']['chronological_decimal_age']
            last_parameter_pair_time_elapsed = last['measurement_dates']['chronological_decimal_age'] - \
                penultimate['measurement_dates']['chronological_decimal_age']
            first_parameter_pair_difference = 0.0
            last_parameter_pair_difference = 0.0

            last_parameter_pair_difference = last['child_observation_value']['observation_value'] - \
                penultimate['child_observation_value']['observation_value']
            first_parameter_pair_difference = penultimate['child_observation_value']['observation_value'] - \
                antepentultimate['child_observation_value']['observation_value']

            latest_velocity = last_parameter_pair_difference / last_parameter_pair_time_elapsed
            penultimate_velocity = first_parameter_pair_difference / \
                first_parameter_pair_time_elapsed
            acceleration = (latest_velocity - penultimate_velocity) / \
                last_parameter_pair_time_elapsed
            return acceleration

def create_pairs(measurements_array: list = []):
    # test data
    """
    measurements = [{'birth_data': {'birth_date': 'Wed, 11 Apr 1759 00:00:00 GMT', 'gestation_weeks': 0, 'gestation_days': 0, 'estimated_date_delivery': 'None', 'estimated_date_delivery_string': '', 'sex': 'female'}, 'measurement_dates': {'observation_date': 'Wed, 25 Apr 1759 00:00:00 GMT', 'chronological_decimal_age': 0.038329911019849415, 'corrected_decimal_age': 0.038329911019849415, 'chronological_calendar_age': '2 weeks', 'corrected_calendar_age': '', 'corrected_gestational_age': {'corrected_gestation_weeks': None, 'corrected_gestation_days': None}, 'clinician_decimal_age_comment': 'Correction for gestational age is nolonger necessary after two years of age.', 'lay_decimal_age_comment': 'Your child is now old enough nolonger to need to take their prematurity into account when considering their growth.'}, 'child_measurement_value': {'height': None, 'weight': 4.111224921050807, 'bmi': 'None', 'ofc': None}, 'measurement_calculated_values': {'height_sds': 'None', 'height_centile': 'None', 'clinician_height_comment': '', 'lay_height_comment': '', 'weight_sds': 1.0020115566532506, 'weight_centile': 84.18309943393204, 'clinician_weight_comment': '', 'lay_weight_comment': '', 'bmi_sds': 'None', 'bmi_centile': 'None', 'clinician_bmi_comment': '', 'lay_bmi_comment': '', 'ofc_sds': 'None', 'ofc_centile': 'None', 'clinician_ofc_comment': '', 'lay_ofc_comment': ''}}, {'birth_data': {'birth_date': 'Wed, 11 Apr 1759 00:00:00 GMT', 'gestation_weeks': 0, 'gestation_days': 0, 'estimated_date_delivery': 'None', 'estimated_date_delivery_string': '', 'sex': 'female'}, 'measurement_dates': {'observation_date': 'Wed, 09 May 1759 00:00:00 GMT', 'chronological_decimal_age': 0.07665982203969883, 'corrected_decimal_age': 0.07665982203969883, 'chronological_calendar_age': '4 weeks', 'corrected_calendar_age': '', 'corrected_gestational_age': {'corrected_gestation_weeks': None, 'corrected_gestation_days': None}, 'clinician_decimal_age_comment': 'Correction for gestational age is nolonger necessary after two years of age.', 'lay_decimal_age_comment': 'Your child is now old enough nolonger to need to take their prematurity into account when considering their growth.'}, 'child_measurement_value': {'height': None, 'weight': 4.699038339425533, 'bmi': 'None', 'ofc': None}, 'measurement_calculated_values': {'height_sds': 'None', 'height_centile': 'None', 'clinician_height_comment': '', 'lay_height_comment': '', 'weight_sds': 1.002339620693291, 'weight_centile': 84.19102035307033, 'clinician_weight_comment': '', 'lay_weight_comment': '', 'bmi_sds': 'None', 'bmi_centile': 'None', 'clinician_bmi_comment': '', 'lay_bmi_comment': '', 'ofc_sds': 'None', 'ofc_centile': 'None', 'clinician_ofc_comment': '', 'lay_ofc_comment': ''}}, {'birth_data': {'birth_date': 'Wed, 11 Apr 1759 00:00:00 GMT', 'gestation_weeks': 0, 'gestation_days': 0, 'estimated_date_delivery': 'None', 'estimated_date_delivery_string': '', 'sex': 'female'}, 'measurement_dates': {'observation_date': 'Wed, 23 May 1759 00:00:00 GMT', 'chronological_decimal_age': 0.11498973305954825, 'corrected_decimal_age': 0.11498973305954825, 'chronological_calendar_age': '1 month, 1 week and 5 days', 'corrected_calendar_age': '', 'corrected_gestational_age': {'corrected_gestation_weeks': None, 'corrected_gestation_days': None}, 'clinician_decimal_age_comment': 'Correction for gestational age is nolonger necessary after two years of age.', 'lay_decimal_age_comment': 'Your child is now old enough nolonger to need to take their prematurity into account when considering their growth.'}, 'child_measurement_value': {'height': None, 'weight': 5.232213641808542, 'bmi': 'None', 'ofc': None}, 'measurement_calculated_values': {'height_sds': 'None', 'height_centile': 'None', 'clinician_height_comment': '', 'lay_height_comment': '', 'weight_sds': 1.0045936184500082, 'weight_centile': 84.24537143099158, 'clinician_weight_comment': '', 'lay_weight_comment': '', 'bmi_sds': 'None', 'bmi_centile': 'None', 'clinician_bmi_comment': '', 'lay_bmi_comment': '', 'ofc_sds': 'None', 'ofc_centile': 'None', 'clinician_ofc_comment': '', 'lay_ofc_comment': ''}}, {'birth_data': {'birth_date': 'Wed, 11 Apr 1759 00:00:00 GMT', 'gestation_weeks': 0, 'gestation_days': 0, 'estimated_date_delivery': 'None', 'estimated_date_delivery_string': '', 'sex': 'female'}, 'measurement_dates': {'observation_date': 'Wed, 06 Jun 1759 00:00:00 GMT', 'chronological_decimal_age': 0.15331964407939766, 'corrected_decimal_age': 0.15331964407939766, 'chronological_calendar_age': '1 month, 3 weeks and 5 days', 'corrected_calendar_age': '', 'corrected_gestational_age': {'corrected_gestation_weeks': None, 'corrected_gestation_days': None}, 'clinician_decimal_age_comment': 'Correction for gestational age is nolonger necessary after two years of age.', 'lay_decimal_age_comment': 'Your child is now old enough nolonger to need to take their prematurity into account when considering their growth.'}, 'child_measurement_value': {'height': None, 'weight': 5.692927141647227, 'bmi': 'None', 'ofc': None}, 'measurement_calculated_values': {'height_sds': 'None', 'height_centile': 'None', 'clinician_height_comment': '', 'lay_height_comment': '', 'weight_sds': 1.004963324529577, 'weight_centile': 84.25427448883711, 'clinician_weight_comment': '', 'lay_weight_comment': '', 'bmi_sds': 'None', 'bmi_centile': 'None', 'clinician_bmi_comment': '', 'lay_bmi_comment': '', 'ofc_sds': 'None', 'ofc_centile': 'None', 'clinician_ofc_comment': '', 'lay_ofc_comment': ''}}, {'birth_data': {'birth_date': 'Wed, 11 Apr 1759 00:00:00 GMT', 'gestation_weeks': 0, 'gestation_days': 0, 'estimated_date_delivery': 'None', 'estimated_date_delivery_string': '', 'sex': 'female'}, 'measurement_dates': {'observation_date': 'Wed, 20 Jun 1759 00:00:00 GMT', 'chronological_decimal_age': 0.19164955509924708, 'corrected_decimal_age': 0.19164955509924708, 'chronological_calendar_age': '2 months, 1 week and 2 days', 'corrected_calendar_age': '', 'corrected_gestational_age': {'corrected_gestation_weeks': None, 'corrected_gestation_days': None}, 'clinician_decimal_age_comment': 'Correction for gestational age is nolonger necessary after two years of age.', 'lay_decimal_age_comment': 'Your child is now old enough nolonger to need to take their prematurity into account when considering their growth.'}, 'child_measurement_value': {'height': None, 'weight': 6.099306464415924, 'bmi': 'None', 'ofc': None}, 'measurement_calculated_values': {'height_sds': 'None', 'height_centile': 'None', 'clinician_height_comment': '', 'lay_height_comment': '', 'weight_sds': 1.007109988748447, 'weight_centile': 84.30590392040097, 'clinician_weight_comment': '', 'lay_weight_comment': '', 'bmi_sds': 'None', 'bmi_centile': 'None', 'clinician_bmi_comment': '', 'lay_bmi_comment': '', 'ofc_sds': 'None', 'ofc_centile': 'None', 'clinician_ofc_comment': '', 'lay_ofc_comment': ''}}, {'birth_data': {'birth_date': 'Wed, 11 Apr 1759 00:00:00 GMT', 'gestation_weeks': 0, 'gestation_days': 0, 'estimated_date_delivery': 'None', 'estimated_date_delivery_string': '', 'sex': 'female'}, 'measurement_dates': {'observation_date': 'Wed, 04 Jul 1759 00:00:00 GMT', 'chronological_decimal_age': 0.2299794661190965, 'corrected_decimal_age': 0.2299794661190965, 'chronological_calendar_age': '2 months, 3 weeks and 2 days', 'corrected_calendar_age': '', 'corrected_gestational_age': {'corrected_gestation_weeks': None, 'corrected_gestation_days': None}, 'clinician_decimal_age_comment': 'Correction for gestational age is nolonger necessary after two years of age.', 'lay_decimal_age_comment': 'Your child is now old enough nolonger to need to take their prematurity into account when considering their growth.'}, 'child_measurement_value': {'height': None, 'weight': 6.461379695081464, 'bmi': 'None', 'ofc': None}, 'measurement_calculated_values': {'height_sds': 'None', 'height_centile': 'None', 'clinician_height_comment': '', 'lay_height_comment': '', 'weight_sds': 1.007475030461087, 'weight_centile': 84.31467244800477, 'clinician_weight_comment': '', 'lay_weight_comment': '', 'bmi_sds': 'None', 'bmi_centile': 'None', 'clinician_bmi_comment': '', 'lay_bmi_comment': '', 'ofc_sds': 'None', 'ofc_centile': 'None', 'clinician_ofc_comment': '', 'lay_ofc_comment': ''}}, {'birth_data': {'birth_date': 'Wed, 11 Apr 1759 00:00:00 GMT', 'gestation_weeks': 0, 'gestation_days': 0, 'estimated_date_delivery': 'None', 'estimated_date_delivery_string': '', 'sex': 'female'}, 'measurement_dates': {'observation_date': 'Wed, 18 Jul 1759 00:00:00 GMT', 'chronological_decimal_age': 0.2683093771389459, 'corrected_decimal_age': 0.2683093771389459, 'chronological_calendar_age': '3 months and 1 week', 'corrected_calendar_age': '', 'corrected_gestational_age': {'corrected_gestation_weeks': None, 'corrected_gestation_days': None}, 'clinician_decimal_age_comment': 'Correction for gestational age is nolonger necessary after two years of age.', 'lay_decimal_age_comment': 'Your child is now old enough nolonger to need to take their prematurity into account when considering their growth.'}, 'child_measurement_value': {'height': None, 'weight': 6.789203770743711, 'bmi': 'None', 'ofc': None}, 'measurement_calculated_values': {'height_sds': 'None', 'height_centile': 'None', 'clinician_height_comment': '', 'lay_height_comment': '', 'weight_sds': 1.009650708766674, 'weight_centile': 84.36686671210703, 'clinician_weight_comment': '', 'lay_weight_comment': '', 'bmi_sds': 'None', 'bmi_centile': 'None', 'clinician_bmi_comment': '', 'lay_bmi_comment': '', 'ofc_sds': 'None', 'ofc_centile': 'None', 'clinician_ofc_comment': '', 'lay_ofc_comment': ''}}, {'birth_data': {'birth_date': 'Wed, 11 Apr 1759 00:00:00 GMT', 'gestation_weeks': 0, 'gestation_days': 0, 'estimated_date_delivery': 'None', 'estimated_date_delivery_string': '', 'sex': 'female'}, 'measurement_dates': {'observation_date': 'Wed, 01 Aug 1759 00:00:00 GMT', 'chronological_decimal_age': 0.3066392881587953, 'corrected_decimal_age': 0.3066392881587953, 'chronological_calendar_age': '3 months and 3 weeks', 'corrected_calendar_age': '', 'corrected_gestational_age': {'corrected_gestation_weeks': None, 'corrected_gestation_days': None}, 'clinician_decimal_age_comment': 'Correction for gestational age is nolonger necessary after two years of age.', 'lay_decimal_age_comment': 'Your child is now old enough nolonger to need to take their prematurity into account when considering their growth.'}, 'child_measurement_value': {'height': None, 'weight': 7.088501064489439, 'bmi': 'None', 'ofc': None}, 'measurement_calculated_values': {'height_sds': 'None', 'height_centile': 'None', 'clinician_height_comment': '', 'lay_height_comment': '', 'weight_sds': 1.0108086749449121, 'weight_centile': 84.39459948388198, 'clinician_weight_comment': '', 'lay_weight_comment': '', 'bmi_sds': 'None', 'bmi_centile': 'None', 'clinician_bmi_comment': '', 'lay_bmi_comment': '', 'ofc_sds': 'None', 'ofc_centile': 'None', 'clinician_ofc_comment': '', 'lay_ofc_comment': ''}}, {'birth_data': {'birth_date': 'Wed, 11 Apr 1759 00:00:00 GMT', 'gestation_weeks': 0, 'gestation_days': 0, 'estimated_date_delivery': 'None', 'estimated_date_delivery_string': '', 'sex': 'female'}, 'measurement_dates': {'observation_date': 'Wed, 15 Aug 1759 00:00:00 GMT', 'chronological_decimal_age': 0.34496919917864477, 'corrected_decimal_age': 0.34496919917864477, 'chronological_calendar_age': '4 months', 'corrected_calendar_age': '', 'corrected_gestational_age': {'corrected_gestation_weeks': None, 'corrected_gestation_days': None}, 'clinician_decimal_age_comment': 'Correction for gestational age is nolonger necessary after two years of age.', 'lay_decimal_age_comment': 'Your child is now old enough nolonger to need to take their prematurity into account when considering their growth.'}, 'child_measurement_value': {'height': None, 'weight': 7.363720070832382, 'bmi': 'None', 'ofc': None}, 'measurement_calculated_values': {'height_sds': 'None', 'height_centile': 'None', 'clinician_height_comment': '', 'lay_height_comment': '', 'weight_sds': 1.0126387519444007, 'weight_centile': 84.4383628580822, 'clinician_weight_comment': '', 'lay_weight_comment': '', 'bmi_sds': 'None', 'bmi_centile': 'None', 'clinician_bmi_comment': '', 'lay_bmi_comment': '', 'ofc_sds': 'None', 'ofc_centile': 'None', 'clinician_ofc_comment': '', 'lay_ofc_comment': ''}}, {'birth_data': {'birth_date': 'Wed, 11 Apr 1759 00:00:00 GMT', 'gestation_weeks': 0, 'gestation_days': 0, 'estimated_date_delivery': 'None', 'estimated_date_delivery_string': '', 'sex': 'female'}, 'measurement_dates': {'observation_date': 'Wed, 29 Aug 1759 00:00:00 GMT', 'chronological_decimal_age': 0.38329911019849416, 'corrected_decimal_age': 0.38329911019849416, 'chronological_calendar_age': '4 months, 2 weeks and 4 days', 'corrected_calendar_age': '', 'corrected_gestational_age': {'corrected_gestation_weeks': None, 'corrected_gestation_days': None}, 'clinician_decimal_age_comment': 'Correction for gestational age is nolonger necessary after two years of age.', 'lay_decimal_age_comment': 'Your child is now old enough nolonger to need to take their prematurity into account when considering their growth.'}, 'child_measurement_value': {'height': None, 'weight': 7.614522747033926, 'bmi': 'None', 'ofc': None}, 'measurement_calculated_values': {'height_sds': 'None', 'height_centile': 'None', 'clinician_height_comment': '', 'lay_height_comment': '', 'weight_sds': 1.0139857964039922, 'weight_centile': 84.47052350872625, 'clinician_weight_comment': '', 'lay_weight_comment': '', 'bmi_sds': 'None', 'bmi_centile': 'None', 'clinician_bmi_comment': '', 'lay_bmi_comment': '', 'ofc_sds': 'None', 'ofc_centile': 'None', 'clinician_ofc_comment': '', 'lay_ofc_comment': ''}}]
    """
    parameter_list = []
    if len(measurements_array) < 2:
        return 'Not enough data'
    else:
        for measurement in measurements_array:
            if measurement:
                if measurement['child_measurement_value']['weight'] is not None:
                    parameter_list.append(measurement)

        if len(parameter_list) < 2:
            return f"There are not enough weight values to calculate a velocity."
        else:
            last = parameter_list[-1]
            penultimate = parameter_list[-2]
            penultimate_weight_sds_value = penultimate['measurement_calculated_values']['weight_sds']
            penultimate_decimal_age = penultimate['measurement_dates']['chronological_decimal_age']
            last_weight_sds_value = last['measurement_calculated_values']['weight_sds']
            last_decimal_age = last['measurement_dates']['chronological_decimal_age']


def correlate_weight(penultimate_decimal_age: float, penultimate_sds: float, last_decimal_age: float, last_sds: float):
    """
    Weight velocity of the individual child cannot be predicted without comparison against reference data velocity
    since weight velocity is age dependent. This uses a uses a correlation matrix to look up values against which
    to compare the rate at which the child is gaining or losing weight.
    The formula for conditional weight gain is: (z2 – r x z1) / √1-r^2
    Conditional reference charts to assess weight gain in British infants, T J Cole, Archives of Disease in Childhood 1995; 73: 8-16
    """

    # import the reference
    cwd = os.path.dirname(__file__)  # current location
    file_path = os.path.join(
        cwd, './data_tables/uk-who_resources/RCPCH weight correlation matrix by month.csv')
    data_frame = pd.read_csv(file_path)

    # look up age
    # this is the formula: (z2 – r x z1) / √1-r^2 where z2 is the most recent value and z1 the penultimate
    # and r the correlation

    # rows are penultimate age, columns are last age. to get r we must look up the ages on
    # the correlation matrix
    z2 = last_sds
    z1 = penultimate_sds

    penultimate_months = penultimate_decimal_age * 12
    last_months = last_decimal_age * 12

    if (last_months > 12):
        raise Exception("Cannot calculate values beyond 12 mths.")
    
    """
        bilinear interpolation will be used
            xx      xx
                x
            xx      xx
                        last
                    lower upper
    penultimate     lower    c1    c3
            upper    c2    c4
    
    """

    # indices of the last values
    last_months_lower_index = math.floor(last_months)
    last_months_upper_index = last_months_lower_index + 1

    # indices of the penultimate values
    penultimate_months_lower_index = math.floor(penultimate_months)
    penultimate_months_upper_index = penultimate_months_lower_index + 1

    # bilinear interpolation
    penultimate_lower_lower = data_frame.iloc[penultimate_months_lower_index, last_months_lower_index]
    penultimate_upper_lower = data_frame.iloc[penultimate_months_upper_index, last_months_lower_index]
    intermediate_interpolated_penultimate = interp1d([penultimate_months, penultimate_months+1],[penultimate_lower_lower, penultimate_upper_lower])
    interpolated_penultimate_lower = intermediate_interpolated_penultimate(penultimate_months)

    last_upper_lower = data_frame.iloc[penultimate_months_lower_index, last_months_upper_index]
    last_upper_upper = data_frame.iloc[penultimate_months_upper_index, last_months_upper_index]
    intermediate_interpolated_last = interp1d([penultimate_months_lower_index, penultimate_months_upper_index], [last_upper_lower, last_upper_upper])
    interpolated_penultimate_upper = intermediate_interpolated_last(penultimate_months)

    intermediate_interpolated_last = interp1d([last_months_lower_index, last_months_upper_index],[interpolated_penultimate_lower, interpolated_penultimate_upper])
    r = intermediate_interpolated_last(last_months)

    r = ferret(z1, z2, r)

    print(r)




def ferret(z1, z2, r):
    """
    The formula for conditional weight gain is: (z2 – r x z1) / √1-r^2
    Conditional reference charts to assess weight gain in British infants, T J Cole, Archives of Disease in Childhood 1995; 73: 8-16f
    """
    conditional_weight_gain = (z2 - (z1 * r)) / math.sqrt(1 - pow(r, 2))
    return conditional_weight_gain

def find_nearest_index(array, value):
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return idx

def reverse_ferret(z1, r, Z):
    return z1 * r + Z * math.sqrt(1 - r^2)

def create_thrive_lines():
    # import the reference
    cwd = os.path.dirname(__file__)  # current location
    file_path = os.path.join(
        cwd, './data_tables/uk-who_resources/RCPCH weight correlation matrix by month.csv')
    data_frame = pd.read_csv(file_path)
    parameter_list = []
    years=[]
    months=[]
    year = 0
    month = 0
    while year < 1:
        years.append(years)
        months.append(month)
        year+=7/365.25 # add a week
        month=year*12
    for i in range(0, len(months), 2):
        # get months in pairs
        first_correlation = correlate_weight(i, 0, i+1, 0)
        weight = reverse_ferret()


def create_correlation_surface():
    # import the reference
    cwd = os.path.dirname(__file__)  # current location
    file_path = os.path.join(
        cwd, './data_tables/uk-who_resources/RCPCH weight correlation matrix by month.csv')
    data_frame = pd.read_csv(file_path)
    # fig = plt.figure()
    # ax = plt.axes(projection="3d")
    # index = data_frame.index[data_frame[]]
    # plt.matshow(data_frame.corr())
    # cb = plt.colorbar()
    # cb.ax.tick_params(labelsize=14)
    # plt.title('Correlation Matrix', fontsize=16)
    # plt.show()



"""
These constants are intended to be used throughout both the rcpchgrowth module
and the API for validation of input values and rejection of very high or low values
"""

# HEIGHT CONSTANTS
MINIMUM_LENGTH_CM = 30
MAXIMUM_HEIGHT_CM = 300
# The tallest person for whom there is irrefutable evidence is
# Robert Pershing Wadlow who was 2.72m (8 ft 11.1 in) tall at his death in 1940
# https://en.wikipedia.org/wiki/Robert_Wadlow

# WEIGHT CONSTANTS
MINIMUM_WEIGHT_KG = 0.2
MAXIMUM_WEIGHT_KG = 500
# The heaviest man according to google is Jon Brower Minnoch (US)
# who had suffered from obesity since childhood. In September 1976,
# he measured 185 cm (6 ft 1 in) tall and weighed 442 kg (974 lb; 69 st 9 lb)
# https://en.wikipedia.org/wiki/Jon_Brower_Minnoch

# OFC CONSTANTS
MINIMUM_OFC_CM = 5
MAXIMUM_OFC_CM = 100
# We couldn't find a good answer for the 'largest human head' that we could use
# Instead have presumed that 100 is a good maximum, given that pathological states
# causing very large head size (eg hydrocephalus) render the OFC meaningless for
# growth measurement and assessment purposes.

# BMI CONSTANTS
MINIMUM_BMI_KGM2 = 7.5
MAXIMUM_BMI_KGM2 = 105
# Again, John Brower Minnoch comes to our rescue here. HIs body mass index
# at the time of his death was 105 kg/m2. In terms of the lowest body mass index ever recorded
# there was nly limited web search information that suggested 7.5 kg.m2 as the lowest ever.

# GESTATION CONSTANTS
MINIMUM_GESTATION_WEEKS = 22
MINIMUM_GESTATION_WEEKS_ERROR = f'It is very unlikely that a gestational age of less than {MINIMUM_GESTATION_WEEKS} weeks is correct'
MAXIMUM_GESTATION_WEEKS = 44
MAXIMUM_GESTATION_WEEKS_ERROR = f'It is very unlikely that a gestational age of more than {MAXIMUM_GESTATION_WEEKS} weeks is correct'



# Recommendations from Project board for reporting Centiles

# Lower limit	Upper limit	Centile band	Weight,  Height, Head	BMI
# 	<-6 		Probable error	Probable error
# -6.00	-2.84	Below 0.4th	Below normal range	Very thin
# -2.84	-2.50	0.4th 		Low BMI
# -2.50	-2.17	0.4th-2nd		Low BMI
# -2.17	-1.83	2nd		
# -1.83	-1.50	2nd -9th		
# -1.50	-1.16	9th		
# -1.16	-0.84	9th-25th		
# -0.84	-0.50	25th		
# -0.50	-0.17	25-50th 		
# -0.17	0.17	50th 		
# 0.17	0.50	50-75th 		
# 0.50	0.84	75th 		
# 0.84	1.16	75-91st 		
# 1.16	1.50	91st 		
# 1.50	1.83	91-98th 		Overweight
# 1.83	2.17	98th 		Overweight
# 2.17	2.50	98-99.6th 		Very overweight (obese)
# 2.50	2.84	99.6th 		Very overweight (obese)
# 2.84	6.00	Above 99.6th	Above normal range	Severely obese
# 	>6.00		Probable error	Probable error

def centile_band_for_centile(sds: float, measurement_method: str)->str:
    # this function returns a centile band into which the sds falls
    # params: accepts a sds: float
    # params: accepts a measurement_method as string

    centile_band = ""

    if measurement_method == "bmi":
        measurement_method = "body mass index"
    if measurement_method == "ofc":
        measurement_method = "head circumference"

    if sds <= -6:
        centile_band = "This " + measurement_method + " measurement is well below the normal range. Please review its accuracy."
    elif sds <= -2.84:
        centile_band = "This " + measurement_method + " measurement is below the normal range."
    elif sds <= -2.5:
        centile_band = "This " + measurement_method + " measurement is on or near the 0.4th centile."
    elif sds <= -2.17:
        centile_band = "This " + measurement_method + " measurement is between the 0.4th and 2nd centiles."
    elif sds <= -1.83:
        centile_band = "This " + measurement_method + " measurement is on or near the 2nd centile."
    elif sds <= -1.5:
        centile_band = "This " + measurement_method + " measurement is between the 2nd and 9th centiles."
    elif sds <= -1.16:
        centile_band = "This " + measurement_method + " measurement is on or near the 9th centile."
    elif sds <= -0.84:
        centile_band = "This " + measurement_method + " measurement is between the 9th and 25th centiles."
    elif sds <= -0.5:
        centile_band = "This " + measurement_method + " measurement is on or near the 25th centile."
    elif sds <= -0.17:
        centile_band = "This " + measurement_method + " measurement is between the 25th and 50th centiles."
    elif sds <= 0.17:
        centile_band = "This " + measurement_method + " measurement is on or near the 50th centile."
    elif sds <= 0.5:
        centile_band = "This " + measurement_method + " measurement is between the 50th and 75th centiles."
    elif sds <= 0.84:
        centile_band = "This " + measurement_method + " measurement is on or near the 75th centile."
    elif sds <= 1.16:
        centile_band = "This " + measurement_method + " measurement is between the 75th and 91st centiles."
    elif sds <= 1.5:
        centile_band = "This " + measurement_method + " measurement is on or near the 91st centile."
    elif sds <= 1.83:
        centile_band = "This " + measurement_method + " measurement is between the 91st and 98th centiles."
    elif sds <= 2.17:
        centile_band = "This " + measurement_method + " measurement is on or near the 98th centile."
    elif sds <= 2.5:
        centile_band = "This " + measurement_method + " measurement is between the 98th and 99.6th centiles."
    elif sds <= 2.84:
        centile_band = "This " + measurement_method + " measurement is on or near the 99.6th centile."
    elif sds <= 6:
        centile_band = "This " + measurement_method + " measurement is above the normal range."
    else:
        centile_band = "This " + measurement_method + " measurement is well above the normal range. Please review its accuracy."

    return centile_band



"""
This file defines the schema for a Measurement object, enabling validation
"""

from marshmallow import Schema, fields, validate


class MeasurementClassSchema(Schema):
    """
    Marshmallow schema for Measurement class
    """
    sex = fields.String(
        required=True,
        validate=validate.OneOf(['male', 'female']),
        description="The sex of the patient, as a string value which can either be `male` or `female`. Abbreviations or alternatives are not accepted")

    birth_date = fields.Date(
        required=True,
        description="Date of birth of the patient, as a Python Date object.")

    observation_date = fields.Date(
        required=True,
        description="The date that the measurement was taken, as a Python Date object.")

    measurement_method = fields.String(
        required=True,
        validate=validate.OneOf(["height", "weight", "bmi", "ofc"]),
        description="The type of measurement performed on the infant or child as a string which can be `height`, `weight`, `bmi` or `ofc`. The value of this measurement is supplied as the `observation_value` parameter. The measurements represent height **in centimetres**, weight *in kilograms**, body mass index **in kilograms/metre²** and occipitofrontal circumference (head circumference, OFC) **in centimetres**.")

    observation_value = fields.Float(
        required=True,
        description="The value of the measurement supplied. This is supplied as a floating point Python number")

    reference = fields.String(
        required=True,
        validate=validate.OneOf(["uk-who", "turners-syndrome", "trisomy-21"]),
        description="The set of references to use for calculation of all the parameters. This can is selected by passing a string which can be 'uk-who', 'turners-syndrome', or 'trisomy-21'. For most UK applications, 'uk-who' is the correct reference. 'turners-syndrome' and 'trisomy-21' are specialist references for use in patients with specific medical conditions, and are more likely to be used only in a specialist context. Use of the incorrect reference will result in totally incorrect data, please ensure you know which data set you wish to use. Contact the RCPCH Growth Charts team if you need help or want to add a new reference")

    gestation_weeks = fields.Number(
        description="The number of completed weeks of gestation at which the patient was born, passed as an integer. Supplying this data enables Gestational Age correction if the child was not born at term. If no gestational age is passed then term is assumed. IMPORTANT: See also the other parameter `gestation_days` - both are usually required.")

    gestation_days = fields.Number(
        description="The number of additional days _beyond the completed weeks of gestation_ at which the patient was born, passed as an integer. Supplying this data enables Gestational Age correction if the child was not born at term. If no gestational age is passed then term is assumed. IMPORTANT: See also the other parameter `gestation_weeks` - both are usually required.")



def comment_prematurity_correction(
    chronological_decimal_age: float,
    corrected_decimal_age: float,
    gestation_weeks: int,
    gestation_days: int):
    """
    Returns interpretations on age correction as a string
    """

    if chronological_decimal_age == corrected_decimal_age:
        # no adjustment has been made so the baby was born at 40 weeks
            lay_corrected_decimal_age_comment = "Your baby was born on their due date."
            clinician_corrected_decimal_age_comment = "Born at term. No correction has been made for gestation."
            lay_chronological_decimal_age_comment = "Your baby was born on their due date."
            clinician_chronological_decimal_age_comment = "Born Term. No correction has been made for gestation."
    elif chronological_decimal_age > corrected_decimal_age:
        ## adjustment for gestational age has been made - even if >=37 weeks
        lay_corrected_decimal_age_comment = f"Because your child was born at {gestation_weeks}+{gestation_days} weeks gestation, an adjustment has been made to take this into account."
        clinician_corrected_decimal_age_comment = "Correction for gestational age has been made."
        lay_chronological_decimal_age_comment = "This is your child's age without taking into account their gestation at birth."
        clinician_chronological_decimal_age_comment = "No correction has been made for gestational age."
        if gestation_weeks < 23:
            lay_corrected_decimal_age_comment = "Your child has been born below the threshold of the charts."
            clinician_corrected_decimal_age_comment = "Your child has been born below the threshold of the charts."
            lay_chronological_decimal_age_comment = "Your child has been born below the threshold of the charts."
            clinician_chronological_decimal_age_comment = "Your child has been born below the threshold of the charts."
    else:
        #some error
        lay_corrected_decimal_age_comment = "It has not been possible to calculate age this time."
        clinician_corrected_decimal_age_comment = "It has not been possible to calculate age this time."
        lay_chronological_decimal_age_comment = "It has not been possible to calculate age this time."
        clinician_chronological_decimal_age_comment = "It has not been possible to calculate age this time."

    comment = {
        'lay_corrected_comment': lay_corrected_decimal_age_comment,
        'lay_chronological_comment': lay_chronological_decimal_age_comment,
        'clinician_corrected_comment': clinician_corrected_decimal_age_comment,
        'clinician_chronological_comment': clinician_chronological_decimal_age_comment
    }
    return comment


# standard imports
from datetime import datetime
import json
import os
from pprint import pprint

# third-party imports
import pytest

# rcpch imports
from rcpchgrowth import Measurement

# the ACCURACY constant defines the accuracy of the test comparisons
# owing to variations in statistical calculations it's impossible to get exact
# agreement between R and Python, so our statistician feels we can set a tolerance
# within which we will accept a result as correct.

ACCURACY = 1e-3


def load_valid_data_set():
    """
    Loads in the testing data from JSON file
    """
    with open(os.path.abspath(os.path.dirname(__file__)) + "/sds_age_validation_2021.json") as f:
        return json.load(f)


@pytest.mark.parametrize("line", load_valid_data_set())
def test_measurement_class_ukwho_data(line):
    """
    Test which iterates through a JSON file of 4000 fictional children and known calculated (TC via R)
      correct SDS values. Compares the output of the Measurement.measurement method with the known
      correct value.
    """

    measurement_object = Measurement(
        sex=str(line["sex"]),
        birth_date=datetime.strptime(line["birth_date"], "%d/%m/%Y"),
        observation_date=datetime.strptime(
            line["observation_date"], "%d/%m/%Y"),
        measurement_method=str(line["measurement_method"]),
        observation_value=float(line["observation_value"]),
        gestation_weeks=int(line["gestation_weeks"]),
        gestation_days=int(line["gestation_days"]),
        reference="uk-who"
    )

    pprint(vars(measurement_object))
    pprint(line)

    # all comparisons using absolute tolerance (not relative)

    # this conditional guards against failure of pytest.approx with NoneTypes
    if line["corrected_sds"] is None:
        assert measurement_object.measurement[
            "measurement_calculated_values"]['corrected_sds'] is None
    else:
        assert measurement_object.measurement[
            "measurement_calculated_values"]['corrected_sds'] == pytest.approx(
            line["corrected_sds"], abs=ACCURACY)

    # this conditional guards against failure of pytest.approx with NoneTypes
    if line["chronological_sds"] is None:
        assert measurement_object.measurement[
            "measurement_calculated_values"]['chronological_sds'] is None
    else:
        assert measurement_object.measurement[
            "measurement_calculated_values"]['chronological_sds'] == pytest.approx(
            line["chronological_sds"], abs=ACCURACY)



"""
Tests the validation of the inputs to Measurement class at instantiation
"""

# standard imports
from datetime import datetime
import pytest

# third-party imports
from marshmallow import exceptions

# rcpch imports
from rcpchgrowth import Measurement


def test_measurement_class_with_invalid_sex_type():
    """
    Should raise a Marshmallow ValidationError (sex must be a string)
    """
    with pytest.raises(exceptions.ValidationError) as error_info:
        Measurement(
            sex=0,
            birth_date=datetime.strptime("2020-04-01", "%Y-%m-%d"),
            observation_date=datetime.strptime("2020-06-01", "%Y-%m-%d"),
            measurement_method="weight",
            observation_value=5.0,
            gestation_weeks=0,
            gestation_days=40,
            reference="uk-who"
        )
    assert "{'sex': ['Not a valid string.']}" in (str(error_info.value))


def test_measurement_class_with_invalid_sex_string():
    """
    Should raise a ValueError (sex must be "male" OR "female")
    """
    with pytest.raises(exceptions.ValidationError) as error_info:
        Measurement(
            sex="maleS",
            birth_date=datetime.strptime("2020-04-01", "%Y-%m-%d"),
            observation_date=datetime.strptime("2020-06-01", "%Y-%m-%d"),
            measurement_method="weight",
            observation_value=5.0,
            gestation_weeks=0,
            gestation_days=40,
            reference="uk-who"
        )
    assert "{'sex': ['Must be one of: male, female.']}" in (
        str(error_info.value))


def test_measurement_class_with_invalid_birth_date_type():
    """
    Should raise an AttributeError during date -> str conversion (date must be a Python datetime object)
    """
    with pytest.raises(AttributeError) as error_info:
        Measurement(
            sex="male",
            birth_date="Date strings are not valid!",
            observation_date=datetime.strptime("2020-06-01", "%Y-%m-%d"),
            measurement_method="weight",
            observation_value=5.0,
            gestation_weeks=0,
            gestation_days=40,
            reference="uk-who"
        )
    assert "'str' object has no attribute 'strftime'" in str(error_info.value)

def test_measurement_class_with_invalid_observation_date_type():
    """
    Should raise an AttributeError during date -> str conversion (date must be a Python datetime object)
    """
    with pytest.raises(AttributeError) as error_info:
        Measurement(
            sex="male",
            birth_date=datetime.strptime("2020-06-01", "%Y-%m-%d"),
            observation_date="Date strings are not valid!",
            measurement_method="weight",
            observation_value=5.0,
            gestation_weeks=0,
            gestation_days=40,
            reference="uk-who"
        )
    assert "'str' object has no attribute 'strftime'" in str(error_info.value)

def test_measurement_class_with_invalid_birth_data():
    """
    Should raise an (observation_date must be after birth_date)
    """
    m = Measurement(
        sex="male",
        birth_date=datetime.strptime("2020-06-01", "%Y-%m-%d"),
        observation_date=datetime.strptime("2020-05-01", "%Y-%m-%d"),
        measurement_method="weight",
        observation_value=5.0,
        gestation_weeks=0,
        gestation_days=40,
        reference="uk-who"
    ).measurement

    # This assertion may change once issues #155 and #157 are closed
    assert m["measurement_dates"]["chronological_decimal_age_error"] == "Chronological age calculation error."

def test_measurement_class_with_invalid_measurement_method():
    """
    Should raise a ValueError (value should be one of [height, weight, bmi, ofc.])
    """
    with pytest.raises(exceptions.ValidationError) as error_info:
        Measurement(
            sex="male",
            birth_date=datetime.strptime("2020-04-01", "%Y-%m-%d"),
            observation_date=datetime.strptime("2020-06-01", "%Y-%m-%d"),
            measurement_method="length",
            observation_value=5.0,
            gestation_weeks=0,
            gestation_days=40,
            reference="uk-who")

    assert "{'measurement_method': ['Must be one of: height, weight, bmi, ofc.']}" in (str(error_info.value))

def test_measurement_class_with_invalid_measurement_method_type():
    """
    Should raise a ValidationError as measurement_method must be a str
    """
    with pytest.raises(exceptions.ValidationError) as error_info:
        Measurement(
            sex="male",
            birth_date=datetime.strptime("2020-04-01", "%Y-%m-%d"),
            observation_date=datetime.strptime("2020-06-01", "%Y-%m-%d"),
            measurement_method=25,
            observation_value=5.0,
            gestation_weeks=0,
            gestation_days=40,
            reference="uk-who")

    assert "{'measurement_method': ['Not a valid string.']}" in (str(error_info.value))

def test_measurement_class_with_invalid_observation_value_type():
    """
    Should raise a ValidationError as measurement_method must be a float
    """
    with pytest.raises(exceptions.ValidationError) as error_info:
        Measurement(
            sex="male",
            birth_date=datetime.strptime("2020-04-01", "%Y-%m-%d"),
            observation_date=datetime.strptime("2020-06-01", "%Y-%m-%d"),
            measurement_method="height",
            observation_value="tall",
            gestation_weeks=0,
            gestation_days=40,
            reference="uk-who")

    assert "{'observation_value': ['Not a valid number.']}" in (str(error_info.value))


def test_measurement_class_with_invalid_observation_value_weight_high():
    """
    Should raise an (too high value error)
    """
    m = Measurement(
        sex="male",
        birth_date=datetime.strptime("2020-06-01", "%Y-%m-%d"),
        observation_date=datetime.strptime("2020-07-01", "%Y-%m-%d"),
        measurement_method="weight",
        observation_value=1000,
        gestation_weeks=0,
        gestation_days=40,
        reference="uk-who"
    ).measurement

    # This assertion may change once issues #155 and #157 are closed
    assert m["child_observation_value"]["observation_value_error"] == "1000 kilograms is very high. Weight must be passed in kilograms."

def test_measurement_class_with_invalid_observation_value_height_low():
    """
    Should raise an (height should be cm not metres error)
    """
    m = Measurement(
        sex="male",
        birth_date=datetime.strptime("2020-06-01", "%Y-%m-%d"),
        observation_date=datetime.strptime("2020-07-01", "%Y-%m-%d"),
        measurement_method="height",
        observation_value=1,
        gestation_weeks=0,
        gestation_days=40,
        reference="uk-who"
    ).measurement


    assert m["child_observation_value"]["observation_value_error"] == "Height/length must be passed in cm, not metres"



def test_measurement_class_with_invalid_gestation_weeks_type():
    """
    Should raise a ValidationError as gestation_weeks must be a float
    """
    with pytest.raises(exceptions.ValidationError) as error_info:
        Measurement(
            sex="male",
            birth_date=datetime.strptime("2020-04-01", "%Y-%m-%d"),
            observation_date=datetime.strptime("2020-06-01", "%Y-%m-%d"),
            measurement_method="height",
            observation_value=53,
            gestation_weeks="two",
            gestation_days=40,
            reference="uk-who")

    assert "{'gestation_weeks': ['Not a valid number.']}" in (str(error_info.value))

def test_measurement_class_with_invalid_gestation_days_type():
    """
    Should raise a ValidationError as gestation_days must be a float
    """
    with pytest.raises(exceptions.ValidationError) as error_info:
        Measurement(
            sex="male",
            birth_date=datetime.strptime("2020-04-01", "%Y-%m-%d"),
            observation_date=datetime.strptime("2020-06-01", "%Y-%m-%d"),
            measurement_method="height",
            observation_value=53,
            gestation_weeks=40,
            gestation_days="three",
            reference="uk-who")

    assert "{'gestation_days': ['Not a valid number.']}" in (str(error_info.value))

def test_measurement_class_with_invalid_reference_type():
    """
    Should raise a ValidationError as reference must be a str
    """
    with pytest.raises(exceptions.ValidationError) as error_info:
        Measurement(
            sex="male",
            birth_date=datetime.strptime("2020-04-01", "%Y-%m-%d"),
            observation_date=datetime.strptime("2020-06-01", "%Y-%m-%d"),
            measurement_method="height",
            observation_value=53,
            gestation_weeks=40,
            gestation_days=3,
            reference=123)

    assert "{'reference': ['Not a valid string.']}" in (str(error_info.value))


def test_measurement_class_with_invalid_reference():
    """
    Should raise a ValidationError as reference must be one of ["uk-who", "turners-syndrome", "trisomy-21"]
    """
    with pytest.raises(exceptions.ValidationError) as error_info:
        Measurement(
            sex="male",
            birth_date=datetime.strptime("2020-04-01", "%Y-%m-%d"),
            observation_date=datetime.strptime("2020-06-01", "%Y-%m-%d"),
            measurement_method="height",
            observation_value=53,
            gestation_weeks=40,
            gestation_days=3,
            reference="who")

    assert "{'reference': ['Must be one of: uk-who, turners-syndrome, trisomy-21.']}" in (str(error_info.value))


from datetime import date, datetime
from datetime import timedelta
from dateutil import relativedelta
import math
from .constants import TERM_PREGNANCY_LENGTH_DAYS, TERM_LOWER_THRESHOLD_LENGTH_DAYS, EXTREME_PREMATURITY_THRESHOLD_LENGTH_DAYS

"""
5 functions to calculate age related parameters
 - chronological_decimal_age: returns a decimal age from 2 dates (takes birth_date and observation_date)
 - corrected_decimal_age: returns a corrected decimal age accounting for prematurity (takes birth_date: date, observation_date: date, gestation_weeks: int, gestation_days: int, pregnancy_length_day [optional])
 - chronological_calendar_age: returns a calendar age as a string (takes birth_date or estimated_date_delivery and observation_date)
 - estimated_date_delivery: returns estimated date of delivery in a known premature infant (takes birth_date, gestation_weeks, gestation_days, pregnancy_length_days[optional])
"""

def chronological_decimal_age(birth_date: date, observation_date: date) -> float:

    """
    Calculates a decimal age from two dates supplied as raw dates without times.
    Returns value floating point
    :param birth_date: date of birth
    :param observation_date: date observation made
    """

    days_between = observation_date - birth_date
    chronological_decimal_age = days_between.days / 365.25
    return chronological_decimal_age
    

def corrected_decimal_age(birth_date: date, observation_date: date, gestation_weeks: int, gestation_days: int)->float: 
    """
    Corrects for gestational age across the life course, including term.
    Any baby 37-42 weeks returns decimal age of 0.0
    Depends on chronological_decimal_age
    :param birth_date: date of birth
    :param observation_date: date observation made
    :param gestation_weeks: weeks of gestation up to 40
    :param gestation_days: days in excess of weeks
    """

    if birth_date > observation_date:
        raise Exception("Birth date cannot be after the date of observation.")
    
    correction_days = 0
    pregnancy_length_days = TERM_PREGNANCY_LENGTH_DAYS

    if gestation_weeks == 0:
        gestation_weeks = 40
        gestation_days = 0
        
    pregnancy_length_days = (gestation_weeks * 7) + gestation_days

    ## age correction
    correction_days = TERM_PREGNANCY_LENGTH_DAYS - pregnancy_length_days
    edd = birth_date + timedelta(days=correction_days)
    corrected_age =  chronological_decimal_age(edd, observation_date)

    return corrected_age


def chronological_calendar_age(birth_date: date, observation_date: date) -> str:
    
    """
    returns age in years, months, weeks and days: to return a corrected calendar age use passes EDD instead of birth date
    """

    if birth_date > observation_date:
        raise Exception("Birth date cannot be after the date of observation.")

    difference = relativedelta.relativedelta(observation_date, birth_date)
    years = difference.years
    months = difference.months
    weeks = difference.weeks
    days = difference.days

    date_string = []

    if years == 1:
        date_string.append(str(years) + " year")
    if years > 1:
        date_string.append(str(years) + " years")
    if months > 1:
        date_string.append(str(months) + " months")
    if months == 1:
        date_string.append(str(months) + " month")
    if days == 1:
        date_string.append(str(days) + " day")
    if days > 1:
        if weeks > 0:
            remainingdays = days - (weeks*7)
            if weeks == 1:
                date_string.append(str(weeks) + " week")
            elif weeks > 1:
                date_string.append(str(weeks) + " weeks")
            if remainingdays == 1:
                date_string.append(str(remainingdays) + " day")
            if remainingdays > 1:
                date_string.append(str(remainingdays) + " days")
        else:
            date_string.append(str(days) + " days")
    if len(date_string) > 1:
        return (', '.join(date_string[:-1])) + ' and ' + date_string[-1]
    elif len(date_string) == 1:
        return date_string[0]
    elif birth_date == observation_date:
        return 'Happy Birthday'
    else:
        return ''


def estimated_date_delivery(birth_date: date, gestation_weeks: int, gestation_days: int) -> date:
    """
    Returns estimated date of delivery from gestational age and birthdate
    Will still calculate an estimated date of delivery if already term (>37 weeks)
    """

    pregnancy_length_days = TERM_PREGNANCY_LENGTH_DAYS

    if gestation_weeks > 0:
        pregnancy_length_days = (gestation_weeks * 7) + gestation_days
    
    prematurity = TERM_PREGNANCY_LENGTH_DAYS - pregnancy_length_days

    edd = birth_date + timedelta(days=prematurity)
    return edd


def corrected_gestational_age(birth_date: date, observation_date: date, gestation_weeks: int, gestation_days: int)->str:
    """
    Returns a corrected gestational age
    """
    edd = estimated_date_delivery(birth_date, gestation_weeks, gestation_days)
    forty_two_weeks_gestation_date = edd + timedelta(days=14)

    if observation_date >= forty_two_weeks_gestation_date:
        #beyond 2 weeks post term - chronological age measured in days / weeks / months years
        #no correction
        return {
            "corrected_gestation_weeks": None,
            "corrected_gestation_days": None
        }
    
    pregnancy_length_days = (gestation_weeks * 7) + gestation_days
    time_alive = observation_date - birth_date
    days_of_life = time_alive.days
    days_since_conception = days_of_life + pregnancy_length_days

    corrected_weeks = math.floor(days_since_conception / 7)
    corrected_supplementary_days = days_since_conception - (corrected_weeks * 7)

    if (corrected_weeks == 42 and corrected_supplementary_days > 0) or corrected_weeks > 42:
        # corrected gestational age will not be returned beyond 42 weeks
        corrected_weeks = None
        corrected_supplementary_days = None

    return {
        'corrected_gestation_weeks': corrected_weeks,
        'corrected_gestation_days': corrected_supplementary_days
    }



import json
import pkg_resources
from .constants import *
# from .global_functions import z_score, cubic_interpolation, linear_interpolation, centile, measurement_for_z, nearest_lowest_index, fetch_lms
# import timeit #see below, comment back in if timing functions in this module

"""
birth_date: date of birth
observation_date: date of observation
sex: sex (string, MALE or FEMALE)
decimal_age: chronological, decimal
corrected_age: corrected for prematurity, decimal
measurement_method: height, weight, bmi, ofc (decimal)
observation: value (float)
gestation_weeks: gestational age(weeks), integer
gestation_days: supplementary days of gestation
lms: L, M or S
reference: reference data
"""

#load the reference data

TRISOMY_21_DATA = pkg_resources.resource_filename(__name__, "/data_tables/trisomy_21.json")
with open(TRISOMY_21_DATA) as json_file:
            TRISOMY_21_DATA = json.load(json_file)
            json_file.close()

def reference_data_absent( 
        age: float,
        measurement_method: str,
        sex: str
    ):
    """
    Helper function.
    Returns boolean
    Tests presence of valid reference data for a given measurement request

    Reference data is not complete for all ages/sexes/measurements.
     - There is only BMI reference data until 18.92y
     - Head circumference reference data is available until 18.0y
     - lowest threshold is 0 weeks, upper threshold is 20y
    """

    if age < 0: # lower threshold of trisomy_21 data
        return True, "No reference data exists below 40 weeks gestation"
    
    if age > TWENTY_YEARS: # upper threshold of trisomy_21 data
        return True, "Trisomy 21 reference data does not exist over the age of 20y."
        
    elif measurement_method == "bmi" and age > 18.82:
        return True, f"Trisomy BMI reference data does not exist > 18.82 y."
    
    elif measurement_method == "ofc":
        if age > EIGHTEEN_YEARS:
            return True, "Trisomy head circumference reference data does not exist > 18 y"
        else:
            return False, ""
    else:
        return False, ""

def trisomy_21_lms_array_for_measurement_and_sex(
        measurement_method: str,
        sex: str,
        age: float
    ):

    data_invalid, data_error = reference_data_absent(age=age, measurement_method=measurement_method, sex=sex)

    if data_invalid:
        raise LookupError(data_error)
    else:
         return TRISOMY_21_DATA["measurement"][measurement_method][sex]

def select_reference_data_for_trisomy_21(measurement_method:str, sex:str):
    try:
        return_value = trisomy_21_lms_array_for_measurement_and_sex(measurement_method=measurement_method, sex=sex, age=0.0)
    except:
        raise LookupError(f"No data for {measurement_method} in the {sex} Trisomy 21 dataset.")
    return return_value


TERM_PREGNANCY_LENGTH_DAYS = 40 * 7
TERM_LOWER_THRESHOLD_LENGTH_DAYS = 37 * 7
EXTREME_PREMATURITY_THRESHOLD_LENGTH_DAYS = 32 * 7
TWENTY_FIVE_WEEKS_GESTATION = ((25 * 7) - (40*7)) / 365.25  # 25 weeks as decimal age
THIRTY_SEVEN_WEEKS_GESTATION = ((37 * 7) - (40*7)) / 365.25  # 37 weeks as decimal age
FORTY_TWO_WEEKS_GESTATION = ((42 * 7) - (40*7)) / 365.25  # 42 weeks as decimal age
TWENTY_THREE_WEEKS_GESTATION = ((23 * 7) - (40*7)) / 365.25  # 23 weeks as decimal age
SEVENTEEN_YEARS = 17.0
EIGHTEEN_YEARS = 18.0
TWENTY_YEARS = 20.0
TERM_WEEKS = 40
LOWER_THRESHOLD_PRETERM_REFERENCE_WEEKS = 23
